import os

from django.test import TestCase
from django.test.utils import override_settings
from django.urls import reverse
from document_catalogue import models, permissions
from . import base


class BaseTestWithUsers(TestCase):
    """
        Test behaviours for Document CRUD views
    """
    def setUp(self):
        super().setUp()
        self.categories = base.create_document_categories()
        self.privilegedUser = base.create_user(username='privileged',
                                               permissions=('Can add document', 'Can change document', 'Can delete document'))
        self.restrictedUser = base.create_user(username='restricted')

    def login(self, user):
        response = self.client.login(username=user.username, password='password')
        self.assertTrue(response)


class PermissionsTests(BaseTestWithUsers):
    """
    Test default permissions
    """
    def test_user_can_view_document_catalogue(self):
        self.assertFalse(permissions.user_can_view_document_catalogue(base.anonymous_user()))
        self.assertTrue(permissions.user_can_view_document_catalogue(self.restrictedUser))

    def test_user_can_download_document(self):
        self.assertFalse(permissions.user_can_download_document(base.anonymous_user()))
        self.assertTrue(permissions.user_can_download_document(self.restrictedUser))

    def test_user_can_edit_document(self):
        self.assertFalse(permissions.user_can_edit_document(self.restrictedUser))
        self.assertTrue(permissions.user_can_edit_document(self.privilegedUser))

    def test_user_can_post_document(self):
        self.assertFalse(permissions.user_can_post_document(self.restrictedUser))
        self.assertTrue(permissions.user_can_post_document(self.privilegedUser))

    def test_user_can_delete_document(self):
        self.assertFalse(permissions.user_can_delete_document(self.restrictedUser))
        self.assertTrue(permissions.user_can_delete_document(self.privilegedUser))


class SuccessDocumentViewTests(BaseTestWithUsers) :
    """
        SUCCESS -- test privileged user makes perfectly reasonable requests
    """
    def test_catalogue_list_view(self):
        self.login(self.restrictedUser)   # Any logged-in user can view the catalogue, by default
        url = reverse('document_catalogue:catalogue_list')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200, "Catalogue view returned non-success status code.")

    def test_category_list_view(self):
        self.login(self.restrictedUser)   # Any logged-in user can view the catalogue, by default
        category_slug = self.categories[0].slug
        url = reverse('document_catalogue:category_list', kwargs={'slug': category_slug})
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200, "Category view returned non-success status code.")

    def test_document_detail_view(self):
        self.login(self.restrictedUser)   # Any logged-in user can view docuemnts in the catalogue, by default
        document = base.create_document()
        url = reverse('document_catalogue:document_detail', kwargs={'pk': document.pk})
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200, "Document detail view returned non-success status code.")

    def test_document_download_view(self):
        self.login(self.restrictedUser)   # Any logged-in user can view docuemnts in the catalogue, by default
        document = base.create_document()
        url = reverse('document_catalogue:document_download', kwargs={'pk': document.pk})
        if base.appConfig.settings.USE_PRIVATE_FILES:
            response = self.client.get(url, follow=True)
            self.assertEqual(response.status_code, 200,
                             "Document download view should re-direct and respond with file.")
            self.assertEqual(302, response.redirect_chain[0][1])
            self.assertIn(document.file.url, response.redirect_chain[0][0])
        else:
            # test client doesn't serve media, so just check for the redirect and that the path exists
            response = self.client.get(url, follow=False)
            self.assertEqual(response.status_code, 302,
                             "Document download view should re-direct and respond with file.")
            self.assertTrue(os.path.exists(document.file.path))

    def test_document_edit_view_get(self):
        self.login(self.privilegedUser)   # Only privileged users can edit a document
        document = base.create_document()
        url = reverse('document_catalogue:document_edit', kwargs={'pk': document.pk})
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200, "Get document edit view returned non-success status code.")

    def test_document_edit_view_post_errors(self) :
        self.login(self.privilegedUser)   # Only privileged users can edit a document
        document = base.create_document()

        myfile = base.generate_file('Test.txt', file_type='txt')
        with open(myfile.name, 'r') as file:
            new_title = 'A Brave New World!'
            post_data = {'file': file, 'title': new_title}  # missing required category
            url = reverse('document_catalogue:document_edit', kwargs={'pk': document.pk})
            response = self.client.post(url, data=post_data)
            self.assertEqual(response.status_code, 200, "Post to document edit view returned non-success status code.")
            self.assertIn(b'form-group has-error', response.content, "Invalid post data response does not contain form-errors")
        # Cleanup
        os.remove(myfile.name)

    def test_document_edit_view_post_success(self):
        self.login(self.privilegedUser)   # Only privileged users can edit a document
        document = base.create_document()
        new_title = 'A Brave New World!'

        myfile = base.generate_file('Test.txt', file_type='txt')
        with open(myfile.name, 'r') as file:
            post_data = {'file': file, 'title': new_title,
                         'description': 'new description', 'category': document.category.pk,
                         'is_published': ''}
            url = reverse('document_catalogue:document_edit', kwargs={'pk': document.pk})
            response = self.client.post(url, data=post_data)
            self.assertEqual(response.status_code, 302, "Post Document edit view did not redirect after editing doc.")
            self.assertNotIn(b'form-errors', response.content, "Valid post data response contains form-errors")

        doc = models.Document.objects.get(pk=document.pk)
        self.assertEqual(doc.title, new_title, "Valid post to document edit view did not alter the title")
        # Cleanup
        os.remove(myfile.name)

    def test_document_delete_view_get(self):
        self.login(self.privilegedUser)   # Only privileged users can delete a document
        document = base.create_document()
        url = reverse('document_catalogue:document_delete', kwargs={'pk': document.pk})
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200, "Get Document delete view returned non-success status code.")

    def test_document_delete_view_post(self):
        self.login(self.privilegedUser)   # Only privileged users can delete a document
        document = base.create_document()
        url = reverse('document_catalogue:document_delete', kwargs={'pk': document.pk})
        response = self.client.post(url)
        self.assertEqual(response.status_code, 302, "Post Document delete view did not redirect after deleting doc.")
        with self.assertRaises(models.Document.DoesNotExist, msg="Document still exists after requesting delete URL."):
            models.Document.objects.get(pk=document.pk)


class DeniedDocumentViewTests(BaseTestWithUsers) :
    """
        DENIED -- test under-privileged user or otherwise makes unreasonable requests
    """
    def test_catalogue_list_view(self):
        # Only authenticated users can view the catalogue
        url = reverse('document_catalogue:catalogue_list')
        response = self.client.get(url)
        self.assertEqual(response.status_code, 403, "Catalogue view non-denied status code for anonymous user.")

    def test_category_list_view(self):
        # Only authenticated users can view the catalogue
        category_slug = self.categories[0].slug
        url = reverse('document_catalogue:category_list', kwargs={'slug': category_slug})
        response = self.client.get(url)
        self.assertEqual(response.status_code, 403, "Category view non-denied status code for anonymous user.")

    def test_document_detail_view(self):
        # Only authenticated users can view the catalogue
        document = base.create_document()
        url = reverse('document_catalogue:document_detail', kwargs={'pk': document.pk})
        response = self.client.get(url)
        self.assertEqual(response.status_code, 403, "Document detail view non-denied status code for anonymous user.")

    def test_document_download_view(self):
        # Only authenticated users can download documents
        document = base.create_document()
        url = reverse('document_catalogue:document_download', kwargs={'pk': document.pk})
        response = self.client.get(url)
        self.assertEqual(response.status_code, 403, "Document download view non-denied status code for anonymous user.")

    def test_private_file_download(self):
        # With private-storage, only authenticated users can download a document file
        document = base.create_document()
        if base.appConfig.settings.USE_PRIVATE_FILES:
            response = self.client.get(document.file.url)
            self.assertEqual(response.status_code, 403, "Private file download view not denied for anonymous user.")
        else:
            # test client doesn't serve media, so just check the path exists
            self.assertTrue(os.path.exists(document.file.path))

    def test_document_edit_view(self):
        self.login(self.restrictedUser)   # Only privileged users can edit a document
        document = base.create_document()
        url = reverse('document_catalogue:document_edit', kwargs={'pk': document.pk})
        response = self.client.get(url)
        self.assertEqual(response.status_code, 403, "Document edit view non-denied status code for restricted user.")

    def test_document_delete_view(self):
        self.login(self.restrictedUser)   # Only privileged users can delete a document
        document = base.create_document()
        url = reverse('document_catalogue:document_delete', kwargs={'pk': document.pk})
        response = self.client.get(url)
        self.assertEqual(response.status_code, 403, "Document delete view non-denied status code for restricted user.")


class SuccessAjaxApiTests(BaseTestWithUsers) :
    """
        SUCCESS -- test privileged user makes perfectly reasonable requests
    """
    def test_api_post(self):
        self.login(self.privilegedUser)
        category_slug = self.categories[0].slug
        url = reverse('document_catalogue:api_post', kwargs={'slug': category_slug})
        myfile = base.generate_file('Test.txt', file_type='txt')
        with open(myfile.name, 'r') as file:
            post_data = {'file': file}
            response = self.client.post(url, post_data)
            self.assertEqual(response.status_code, 200, "Request to upload document file returned non-success status code.")
            self.assertEqual(models.Document.objects.count(), 1, "Document count incorrect after file upload.")

        document = models.Document.objects.all().first()
        _, filename = os.path.split(myfile.name)
        self.assertEqual(document.title, filename, "Document has different default name from uploaded file.")
        self.assertIn(category_slug, document.file.path, "Document path does not include category slug.")
        # The filename may get altered when saved, but should contain the name and extension of the original
        name, extension = os.path.splitext(filename)
        self.assertIn(name, document.file.path, "Document path does not contain original uploaded file name.")
        self.assertIn(extension, document.file.path, "Document path does not contain original uploaded file name.")
        # Cleanup
        os.remove(myfile.name)

    def test_api_delete(self):
        self.login(self.privilegedUser)   # Only privileged users can delete a document
        document = base.create_document()
        url = reverse('document_catalogue:api_delete', kwargs={'pk': document.pk})
        response = self.client.delete(url)
        self.assertEqual(response.status_code, 200, "Request to delete document returned non-success status code.")
        with self.assertRaises(models.Document.DoesNotExist, msg="Document still exists after requesting delete URL."):
            models.Document.objects.get(pk=document.pk)


class DeniedAjaxApiTests(BaseTestWithUsers) :
    """
        DENIED -- test unprivileged user or otherwise makes unreasonable requests
    """
    def test_api_post_invalid_file(self):
        self.login(self.privilegedUser)
        category_slug = self.categories[0].slug
        url = reverse('document_catalogue:api_post', kwargs={'slug': category_slug})

        myfile = base.generate_file('Test.html', file_type='html')
        with open(myfile.name, 'r') as file:
            post_data = {'file': file}
            response = self.client.post(url, post_data)
            self.assertEqual(response.status_code, 403, "Invalid filetype did not result in PermissionDenied.")
        # Cleanup
        os.remove(myfile.name)

    def test_api_post(self):
        self.login(self.restrictedUser)   # Only privileged users can delete a document
        category_slug = self.categories[0].slug
        url = reverse('document_catalogue:api_post', kwargs={'slug': category_slug})

        myfile = base.generate_file('Test.txt', file_type='txt')
        with open(myfile.name, 'r') as file:
            post_data = {'file': file}
            response = self.client.post(url, post_data)
            self.assertEqual(response.status_code, 403, "Request to upload file non-denied status code for restricted user.")
        # Cleanup
        os.remove(myfile.name)

    def test_api_delete(self):
        self.login(self.restrictedUser)   # Only privileged users can delete a document
        document = base.create_document()
        url = reverse('document_catalogue:api_delete', kwargs={'pk': document.pk})
        response = self.client.delete(url)
        self.assertEqual(response.status_code, 403, "Request to delete document non-denied status code for restricted user.")



from django.apps import apps
from django.urls import path

from . import views

appConfig = apps.get_app_config('document_catalogue')

app_name = 'document_catalogue'

urlpatterns = [
    path('',
         view=views.DocumentCatalogueListView.as_view(),
         name='catalogue_list'
    ),
    path('category/<slug:slug>/',
         view=views.CategoryDocumentListView.as_view(),
         name='category_list'
         ),
    path('detail/<int:pk>/',
         view=views.DocumentDetailView.as_view(),
         name='document_detail'
    ),
    path('download/<int:pk>/',
         view=views.DocumentDownloadView.as_view(),
         name='document_download'
    ),
    # AJAX API
    path('search/',
         view=views.DocumentAjaxAPI.as_view(),
         name='api_search'
    ),
]

if appConfig.settings.ENABLE_EDIT_URLS:
    urlpatterns += [
        path('edit/<int:pk>/',
             view=views.DocumentEditView.as_view(),
             name='document_edit'
        ),
        path('detail/<int:pk>/delete/',
             view=views.DocumentDeleteView.as_view(),
             name='document_delete'
        ),
        # AJAX API
        path('post/<slug:slug>/',
             view=views.DocumentAjaxAPI.as_view(),
             name='api_post'
        ),
        path('delete/<int:pk>/',
             view=views.DocumentAjaxAPI.as_view(),
             name='api_delete'
        ),
    ]



import os

import django.conf
from django.test import TestCase
from django.core.files.uploadedfile import UploadedFile
from django import forms
from document_catalogue import models
from . import base


# Create tests for models here.

class CategoryTests(TestCase):
    """
        Test basic behaviours for DocumentCategory model
    """
    def setUp(self):
        super().setUp()
        self.categories=base.create_document_categories()

    def test_get_absolute_url(self):
        categories = models.DocumentCategory.objects.all()
        for cat in categories:
            self.assertIn(cat.slug, cat.get_absolute_url(), 'URL for category should contain its slug.')

    def test_has_children(self):
        categories = models.DocumentCategory.objects.filter(parent=None)
        for cat in categories:
            self.assertTrue(cat.has_children(), 'Category reporting no children when it has sub-categories.')

    def test_not_has_children(self):
        cat = models.DocumentCategory.objects.get(slug='sub-category-1b')
        self.assertTrue(not cat.has_children(), 'Category reporting children when it has no sub-categories.')

    def test_document_count(self):
        categories = models.DocumentCategory.objects.all()
        for cat in categories:
            self.assertEqual(cat.get_document_count(), 0, 'Category with no documents returns non-zero get_document_count.')


class DocumentTests(TestCase):
    """
        Test basic behaviours for Document model
    """
    FILENAME = 'myDocument.txt'

    def setUp(self):
        super().setUp()
        self.categories=base.create_document_categories()
        self.document = base.create_document(filename=self.FILENAME, file_type='txt')

    def tearDown(self):
        os.remove(self.document.file.path)

    def test_get_absolute_url(self):
        self.assertIn(str(self.document.pk), self.document.get_absolute_url(), 'URL for document should contain its pk.')

    def test_get_filetype(self):
        filetype = self.document.get_filetype()
        self.assertEqual(filetype, 'txt', 'get_filetype returns incorrect type %s' % filetype)

    def test_document_directory_path(self):
        instance = lambda: None  # a mutable null object
        instance.category = self.categories[0]
        path = models.document_upload_path_callback(instance, self.FILENAME)
        self.assertIn(self.categories[0].slug, path)
        self.assertIn(self.FILENAME, path)

    def test_create_file(self):
        filename = 'slartibartfast.txt'
        file = base.create_document(filename=filename, file_type='txt', user=base.create_user(username='slartibartfast'))
        doc = models.Document.objects.get(pk=file.pk)
        self.assertIn(filename, doc.file.name)
        self.assertIn(filename, doc.file.url)
        # Cleanup
        os.remove(doc.file.path)

    def test_private_storage(self):
        media_root = getattr(django.conf.settings, 'PRIVATE_STORAGE_ROOT') if base.appConfig.USE_PRIVATE_FILES \
                                                                         else django.conf.settings.MEDIA_ROOT
        self.assertIn(media_root, self.document.file.path)
        self.assertIn(base.appConfig.settings.MEDIA_ROOT, self.document.file.path)


class ConstrainedfileFieldTests(TestCase):
    """
        A few basic tests for common validation in both PrivateFileField and ConstrainedFileField
    """
    def setUp(self):
        super().setUp()
        self.categories=base.create_document_categories()

    def test_validation_success(self):
        document = base.create_document(filename='dummy.txt', file_type='txt')
        file_field = document.file.field
        self.assertIsNotNone(file_field.clean(value=document.file, model_instance=document))

    class FileField:
        def __init__(self, field):
            self.field = field
            self.store = self.get_max_upload_size()
        def get_max_upload_size(self):
            return self.field.max_file_size if base.appConfig.USE_PRIVATE_FILES else self.field.max_upload_size
        def set_max_upload_size(self, max):
            self.field.max_upload_size = max
        def restore_max_upload_size(self):
            if base.appConfig.USE_PRIVATE_FILES:
                self.field.max_file_size = self.store
            else:
                self.field.max_upload_size = self.store

    def test_max_upload_size_fail(self):
        document = base.create_document()
        file_field = self.FileField(document.file.field)
        file_size = file_field.get_max_upload_size()-1
        document.file.file = \
            UploadedFile(file=document.file.file, name=document.title, content_type='txt', size=file_size)
        # fudge the max_upload_size to fall below file size.
        file_field.set_max_upload_size(document.file.size - 1)
        with self.assertRaises(forms.ValidationError):
            file_field.field.clean(value=document.file, model_instance=document)
        # Cleanup
        file_field.restore_max_upload_size()
        os.remove(document.file.path)

    def test_content_types_fail(self):
        document = base.create_document(filename='dummy.html', file_type='html')
        file_field = document.file.field
        document.file.file = \
            UploadedFile(file=document.file.file, name=document.title, content_type='html', size=document.file.size)
        with self.assertRaises(forms.ValidationError):
            file_field.clean(value=document.file, model_instance=document)
        # Cleanup
        os.remove(document.file.path)



# Generated by Django 3.0.6 on 2020-05-04 20:02

from django.conf import settings
from django.apps import apps
from django.db import migrations, models
import django.db.models.deletion
import document_catalogue.models
import mptt.fields

appConfig = apps.get_app_config('document_catalogue')

file_field = models.FileField(upload_to=document_catalogue.models.document_upload_path_callback)
if appConfig.settings.USE_PRIVATE_FILES:
    import private_storage.fields
    import private_storage.storage.files
    file_field = private_storage.fields.PrivateFileField(
        storage=private_storage.storage.files.PrivateFileSystemStorage(),
        upload_to=document_catalogue.models.document_upload_path_callback)
else:
    import constrainedfilefield.fields

    file_field = constrainedfilefield.fields.ConstrainedFileField(
        max_length=200,
        upload_to=document_catalogue.models.document_upload_path_callback)


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='DocumentCategory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=64)),
                ('slug', models.SlugField(unique=True)),
                ('description', models.TextField(blank=True)),
                ('lft', models.PositiveIntegerField(editable=False)),
                ('rght', models.PositiveIntegerField(editable=False)),
                ('tree_id', models.PositiveIntegerField(db_index=True, editable=False)),
                ('level', models.PositiveIntegerField(editable=False)),
                ('parent', mptt.fields.TreeForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='children', to='document_catalogue.DocumentCategory')),
            ],
            options={
                'verbose_name_plural': 'Categories',
                'ordering': ['slug'],
            },
        ),
        migrations.CreateModel(
            name='Document',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sort_order', models.SmallIntegerField(default=1)),
                ('creation_date', models.DateTimeField(auto_now_add=True, verbose_name='Created')),
                ('update_date', models.DateTimeField(auto_now=True, verbose_name='Last Modified')),
                ('title', models.CharField(max_length=512)),
                ('description', models.TextField(blank=True, null=True)),
                ('is_published', models.BooleanField(default=False)),
                ('file', file_field),
                ('category', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='document_catalogue.DocumentCategory')),
                ('user', models.ForeignKey(on_delete=models.SET(1), to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ('category', 'sort_order'),
            },
        ),
    ]



from __future__ import unicode_literals
from importlib import import_module

from django.apps import apps
from django import template

appConfig = apps.get_app_config('document_catalogue')

# import Plugin Permissions module
permissions = import_module(appConfig.settings.PERMISSIONS)

register = template.Library()

def has_permission(user, perm):
    check_perm = getattr(permissions, perm, lambda user: False)
    return check_perm(user) if callable(check_perm) else False

@register.filter(name='can')
def can_template_tag(user, permission):
    return has_permission(user, permission)



"""
     Base classes used to setup testing fixtures
"""
from django.conf import settings
from django.apps import apps
from django.core.files.uploadedfile import SimpleUploadedFile
from django.contrib.auth.models import AnonymousUser, User, Permission
from django.utils.text import slugify
from document_catalogue import models

appConfig = apps.get_app_config('document_catalogue')


def anonymous_user():
    return AnonymousUser()


def create_user(username='myUser', permissions=()):
    """
        Factory function to create and return a single user with the given iterable of permissions.
        Common DC permissions:  ('Can add document', 'Can change document', 'Can delete document')
    """
    user = User.objects.create_user(
        first_name=username.capitalize(), last_name='Lastname', email='{username}@example.com'.format(username=username),
        username=username, password='password',
    )
    if permissions:
        permissions = Permission.objects.filter(name__in=permissions)
        user.user_permissions.set(permissions)

    return user


def create_document_categories(category_names=(('Top Level Category 1', (('Sub-Category 1A', ()), ('Sub-Category 1B', ()), )),
                                               ('Top Level Category 2', (('Sub-Category 2A', ()), ))
                                              ),
                               parent=None,
                               description_template='This is the description for {slug}',
                               ):
    """
        Factory function to create and return a list of DocumentCategory objects.
        Hierarchy defined by nesting of 2-tuples in category_names.
    """
    def create_category(name, parent=None):
        slug = slugify(name)
        return models.DocumentCategory.objects.create(name=name, slug=slug, parent=parent,
                                                      description=description_template.format(slug=slug))

    categories = []
    for name, children in category_names:
        cat = create_category(name, parent=parent)
        categories += [cat] + create_document_categories(children, parent=cat)

    return categories


def generate_file(filename, file_type='txt'):
    filename = '{media}{filename}'.format(media=settings.BASE_DIR, filename=filename)

    def write_text_file(filename, content):
        with open(filename, 'wb') as myfile:
            myfile.write(content)
            return myfile

    if file_type == 'txt':
        return write_text_file(filename, b'Hello World')

    if file_type == 'html':
        return write_text_file(filename, b'<!DOCTYPE html><html><head></head><body><p>Hello World</p></body></html>')

    raise Exception("Don't know how to generate file of type %s" % file_type)


def generate_simple_uploaded_file(filename, file_type='txt'):
    if file_type == 'txt':
        return SimpleUploadedFile(filename, b'Hello World')

    if file_type == 'html':
        return SimpleUploadedFile(filename, b'<!DOCTYPE html><html><head></head><body><p>Hello World</p></body></html>')

    raise Exception("Don't know how to generate file of type %s" % file_type)


def create_document(filename='hello.txt', file_type='txt', user=None, category=None):
    document = models.Document.objects.create(
        category=category or models.DocumentCategory.objects.all().first(),
        user=user or create_user(permissions=('Can add document')),
        is_published=True,
        file=generate_simple_uploaded_file(filename, file_type),
    )
    return document



from django.contrib import admin
from mptt.admin import DraggableMPTTAdmin
from .models import DocumentCategory, Document


@admin.register(DocumentCategory)
class CategoryAdmin(DraggableMPTTAdmin):
    list_display = ('tree_actions', 'indented_title')
    list_display_links = ('indented_title',)
    prepopulated_fields = {'slug': ('name',)}


@admin.register((Document))
class DocumentAdmin(admin.ModelAdmin):
    list_display = ('title', 'category', 'is_published', 'creation_date', 'update_date')
    list_editable = ('category', 'is_published')



# Generated by Django 3.0.6 on 2020-05-22 22:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('document_catalogue', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='document',
            name='sort_order',
            field=models.PositiveSmallIntegerField(default=0),
        ),
        migrations.AlterModelOptions(
            name='document',
            options={'ordering': ('sort_order',)},
        ),
    ]



""" Re-usable generic views and view mixins """
from django.template.loader import get_template
from django import http
from django.views import generic


class JSONResponseMixin:
    """
    A mixin that can be used to render a JSON response.
    Taken directly from : https://docs.djangoproject.com/en/2.2/topics/class-based-views/mixins/#more-than-just-html
    """
    def render_to_json_response(self, context, **response_kwargs):
        """
        Returns a JSON response, transforming 'context' to make the payload.
        """
        return http.JsonResponse(
            self.get_data(context),
            **response_kwargs
        )

    def get_data(self, context):
        """
        Returns an object that will be serialized as JSON by json.dumps().
        """
        # Note: This is *EXTREMELY* naive; in reality, you'll need
        # to do much more complex handling to ensure that arbitrary
        # objects -- such as Django model instances or querysets
        # -- can be serialized as JSON.
        return context


class AjaxOnlyViewMixin(JSONResponseMixin, generic.base.ContextMixin, generic.base.View):
    """ A generic View with a text/ajax mimetype that yields a 405 for non-AJAX requests """
    form_errors_template = get_template("document_catalogue/ajax/form_errors.html")

    def dispatch(self, request, *args, **kwargs):
        # Add check for Ajax to normal dispatch logic.
        if not request.is_ajax():
            return self._ajax_only_allowed(request)
        else:
            return super().dispatch(request, *args, **kwargs)

    @staticmethod
    def _ajax_only_allowed(request):
        generic.base.logger.warning('Non-AJAX request Not Allowed (%s): %s', request.method, request.path,
            extra={
                'status_code': 405,
                'request': request
            }
        )
        return http.HttpResponseNotAllowed(('AJAX', ))

    def get(self, request, *args, **kwargs):
        """
            Sub-classes must override this method to do something sensible.
        """
        response_data = {
            'dummy': 'some value',
        }
        return self.render_to_json_response(response_data)

    def _form_errors_context(self, form, strip_tags=False):
        return {
            'success': False,
            'errors' : self.form_errors_template.render(context={'form': form, 'strip_tags': strip_tags})
        }


"""
Dependency Injection Micro-Framework using Plugins

    A Plugin provides some custom behaviour to specific Concrete classes in the class hierarchy
    The Base class of the hierarchy generally applies the plugins by adding PluginManager as a mixin
    Plugins can then be injected into any concrete class in its hierarchy.

"""

from abc import ABCMeta
from itertools import chain
from django.db.models.functions import Lower

#########################
#  Abstract Base Classes - plugin architecture
#########################


class PluginManager:
    """
        A simple, generic plugin manager.
        Provides a clean dependency injection mechanism when used as a Mixin: BaseClass(PluginManager)
        BaseClass implements behaviours for plugins (defined by whatever interface suited to app)
        ConcreteSubclass(BaseClass) can then RegisterPlugins to inject additional behaviours
    """
    plugins = []

    @classmethod
    def add_plugins(cls, plugins):
        cls.plugins = list(chain(cls.plugins, plugins))

    @classmethod
    def apply_plugins(cls, f):
        """ Apply f to each plugin.  f must be a callable that takes a single plugin parameter """
        for plugin in cls.plugins :
            try:
                f(plugin)
            except AttributeError:
                pass


class RegisterPlugins :
    """
        A Decorator for registering a set of plugins to a PluginManager
        E.g.:  @RegisterPlugins(Plugin1(some, parameters), Plugin2())
    """
    def __init__(self, *plugins) :
        self.plugins = plugins

    def __call__(self, decorated_class) :
        decorated_class.add_plugins(self.plugins)

        return decorated_class


#########################
#  Concrete Implementations - document_catalogue plugins
#########################


class ViewPluginManager(PluginManager):
    """ Encapsulates logic specific to applying AbstractViewPlugin plugins.  Intended as View mixin """
    @classmethod
    def plugins_extend_qs(cls, request, qs):
        """ Apply each plugin to the given queryset, in sequence, return resulting queryset """
        for plugin in cls.plugins :
            try:
                qs = plugin.extend_qs(request, qs)
            except AttributeError:
                pass
        return qs

    @classmethod
    def plugins_get_context(cls, request):
        context = {}
        cls.apply_plugins(lambda plugin: context.update(plugin.get_context(request)))
        return context


class AbstractViewPlugin(metaclass=ABCMeta):
    """ Defines the API for a plugin that injects behaviour into a View class """
    def apply(self, request):
        """ Apply the plugin to the given request just prior to dispatching it """
        pass

    def extend_qs(self, request, qs):
        """ Extend, modify, or constrain the base document queryset and return it """
        return qs

    def get_context(self, request):
        """ Return a dictionary to be added to the View's context """
        return {}


class OrderedViewPlugin(AbstractViewPlugin):
    """ Applies ordering to view's queryset based on URL query argument found in request.GET """
    ORDERING_CHOICES = (
        ('default', 'Default'),
        ('date', 'Recently Updated'),
        ('title', 'Title')
    )
    ORDERING_KEYS = tuple(k for k,v in ORDERING_CHOICES)

    ORDERING_EXPRESSION = {   # valid ordering expressions maps query param value to ordering clause
        'date' : '-update_date',
        'title': Lower('title').asc(),
    }

    def __init__(self, query_param='dc_ordering'):
        """ Name of the query parameter used to specify ordering """
        super().__init__()
        self.query_param = query_param

    def get_ordering_key(self, request):
        key = request.GET.get(self.query_param, None)
        return key if key in self.ORDERING_KEYS else None

    def get_ordering(self, request):
        """ Return the order_by experession for the queryset """
        return self.ORDERING_EXPRESSION.get(self.get_ordering_key(request), None)

    def extend_qs(self, request, qs):
        ordering = self.get_ordering(request)
        if ordering:
            qs = qs.order_by(ordering)
        return qs

    def get_context(self, request):
        return {
            self.query_param : self.get_ordering_key(request),
            '{ordering}_choices'.format(ordering=self.query_param) : self.ORDERING_CHOICES,
        }


class SessionOrderedViewPlugin(OrderedViewPlugin):
    """ Applies ordering to view's queryset based on ordering passed in URL query arg and stored in session """

    def get_ordering_key(self, request):
        key = request.session.get(self.query_param, None)
        return key if key in self.ORDERING_KEYS else None

    def apply(self, request):
        """ Apply the plugin to the given request """
        ordering_key = super().get_ordering_key(request)
        if ordering_key:
            request.session[self.query_param] = ordering_key



"""Default settings for the document_catalogue app."""

from django.conf import settings
from django.apps import apps

appConfig = apps.get_app_config('document_catalogue')

# To use django-private-storage as storage backend for supporting documentation files
# If True, pip install django-private-storage and configure settings for private file storage as per docs
# Intended to be set at start of project and not changed - schema and data migration required if this changes!
DOCUMENT_CATALOGUE_USE_PRIVATE_FILES = appConfig.USE_PRIVATE_FILES

# IMPORTANT: this only restricts access to the Catalogue views - if private-files is not used, anyone with thier URL can download them!
DOCUMENT_CATALOGUE_LOGIN_REQUIRED = getattr(settings, 'DOCUMENT_CATALOGUE_LOGIN_REQUIRED', appConfig.LOGIN_REQUIRED_DEFAULT)

# Fine scale permissions (default permissions use DOCUMENT_CATALOGUE_LOGIN_REQUIRED, but that setting may be disabled by custom permissions)
# Value is a dotted path to a permissions module or object with the required permissions functions - see permissions.py
DOCUMENT_CATALOGUE_PERMISSIONS = getattr(settings, 'DOCUMENT_CATALOGUE_PERMISSIONS', 'document_catalogue.permissions')

# Enable document editing URL's and the Ajax document API
# When disabled, documents are managed via the django.admin
# When enabled, be sure to carefully consider the permissions on these views!
DOCUMENT_CATALOGUE_ENABLE_EDIT_URLS = getattr(settings, 'DOCUMENT_CATALOGUE_ENABLE_EDIT_URLS', False)

# Root directory for file uploads to the document catalogue
# If private-files is used, media_root will be a sub-directory of PRIVATE_STORAGE_ROOT
DOCUMENT_CATALOGUE_MEDIA_ROOT = getattr(settings, 'DOCUMENT_CATALOGUE_MEDIA_ROOT', 'documents/')

# Upload file settings - constrain the file content_types and max. filesize for document file uploads.
# None to allow any file type - eliminates dependency on python-magic and libmagic, which are included only if a whitelist is supplied.
#   e.g.  DOCUMENT_CATALOGUE_CONTENT_TYPE_WHITELIST = ('application/pdf', 'image/png', 'image/jpg')
DOCUMENT_CATALOGUE_CONTENT_TYPE_WHITELIST = getattr(settings,'DOCUMENT_CATALOGUE_CONTENT_TYPE_WHITELIST', None)
# max. size in bytes, None for no limit
DOCUMENT_CATALOGUE_MAX_FILESIZE = getattr(settings, 'DOCUMENT_CATALOGUE_MAX_FILESIZE', 10 * 1024 * 1024)

# Plugin Classes used to inject behaviours into standard document list views
# Define plugins by extending the ABC: document_catalogue.plugins.AbstractViewPlugin
# Inject them by customizing this setting
DOCUMENT_CATALOGUE_LIST_VIEW_PLUGINS = getattr(settings, 'DOCUMENT_CATALOGUE_LIST_VIEW_PLUGINS',
                                               ('document_catalogue.plugins.SessionOrderedViewPlugin', ))